package de.mdelab.mlsdm.interpreter;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;

import de.mdelab.expressions.interpreter.callaction.CallActionInterpreter;
import de.mdelab.expressions.interpreter.core.ExpressionInterpreterManager;
import de.mdelab.expressions.interpreter.core.ExpressionsInterpreterException;
import de.mdelab.expressions.interpreter.core.variables.Variable;
import de.mdelab.expressions.interpreter.core.variables.VariablesScope;
import de.mdelab.mlcallactions.CallAction;
import de.mdelab.mlcallactions.CallActionParameter;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.ActivityCallAction;
import de.mdelab.mlsdm.MlsdmPackage;
import de.mdelab.mlsdm.interpreter.facade.MLSDMMetamodelFacadeFactory;
import de.mdelab.mlsdm.interpreter.notifications.MLSDMNotificationEmitter;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.SDMInterpreterConstants;
import de.mdelab.sdm.interpreter.core.notifications.Notifier;

/**
 * The CallActionInterpreter executes CallActions, an expression language
 * without a concrete textual syntax. Its main purpose is to allow calling Java
 * methods and other story diagrams from a story diagram.
 * 
 * @author Stephan Hildebrandt
 * 
 */
public class MLSDMCallActionInterpreter extends CallActionInterpreter {

	private MLSDMNotificationEmitter notificationEmitter;
	
	public MLSDMCallActionInterpreter() {
		notificationEmitter = null;
	}

	@Override
	public void setExpressionInterpreterManager( final ExpressionInterpreterManager<EClassifier, ?, MLExpression> expressionInterpreterManager) {
		super.setExpressionInterpreterManager( expressionInterpreterManager );
		
		if ( expressionInterpreterManager instanceof Notifier ) {
			this.notificationEmitter = (MLSDMNotificationEmitter) ( (Notifier<?, ?, ?, ?, ?, ?, ?, ?, ?>) expressionInterpreterManager ).getNotificationEmitter();
		}
	}
	
	@Override
	protected Variable<EClassifier> evaluateCallAction( final CallAction callAction,
														final EClassifier contextClassifier,
														final Object contextInstance,
														final VariablesScope<EClassifier, ?, MLExpression> variablesScope )
	throws ExpressionsInterpreterException {
		if ( callAction.eClass() == MlsdmPackage.Literals.ACTIVITY_CALL_ACTION) {
			try {
				return this.evaluate((ActivityCallAction) callAction, contextClassifier, contextInstance, variablesScope);
			} catch ( final SDMException ex ) {
				throw new ExpressionsInterpreterException( ex );
			}
		}
		
		return super.evaluateCallAction( callAction, contextClassifier, contextInstance, variablesScope );
	}
	
	@SuppressWarnings("unchecked")
	protected MLSDMInterpreter createMLSDMInterpreter() {
		return new MLSDMInterpreter( 	(ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression>) this.getExpressionInterpreterManager(),
										getNotificationEmitter(),
										MLSDMMetamodelFacadeFactory.INSTANCE );
	}

	protected Variable<EClassifier> evaluate(	final ActivityCallAction callAction,
												final EClassifier contextClassifier,
												final Object contextInstance,
												final VariablesScope<EClassifier, ?, MLExpression> variablesScope )
	throws SDMException, ExpressionsInterpreterException {
		final MLSDMInterpreter sdi = createMLSDMInterpreter();

		/*
		 * Create list of variable bindings.
		 */
		final List<Variable<EClassifier>> parameters = new LinkedList<Variable<EClassifier>>();

		for (final CallActionParameter parameter : callAction.getParameters())
		{
			final Variable<EClassifier> variable = new Variable<EClassifier>(parameter.getName(), parameter.getType(), this.evaluate(
					parameter.getValueExpression(), contextClassifier, contextInstance, variablesScope).getValue());

			parameters.add(variable);
		}

		final Map<String, Variable<EClassifier>> returnValues = sdi.executeActivity(callAction.getActivity(), parameters);

		Variable<EClassifier> returnVariable = null;

		if (callAction.getReturnValueParameter() != null)
		{
			returnVariable = returnValues.get(callAction.getReturnValueParameter().getName());
			returnVariable = new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, returnVariable.getClassifier(),
					returnVariable.getValue());
		}
		else
		{
			returnVariable = new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, EcorePackage.Literals.EJAVA_OBJECT, null);
		}

		return returnVariable;
	}

	public MLSDMNotificationEmitter getNotificationEmitter() {
		return notificationEmitter;
	}
}
