package de.mdelab.mlsdm.interpreter.patternMatcher.patternPartBased;

import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.ExpressionsInterpreterException;
import de.mdelab.expressions.interpreter.core.variables.Variable;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.MlstorypatternsPackage;
import de.mdelab.mlstorypatterns.ModifierEnum;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.mlstorypatterns.StoryPatternObject;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.EMatchType;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.PatternPart;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.PatternPartBasedMatcher;

/**
 * Abstract class of all pattern parts specific for the StoryDiagramEcore
 * metamodel.
 * 
 * @author Stephan Hildebrandt
 * 
 * @param <SPO>
 * @param <SPL>
 */
public abstract class MLSDMPatternPart<SPO extends AbstractStoryPatternObject, SPL extends AbstractStoryPatternLink> extends
		PatternPart<AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>
{
	protected boolean			isCreate	= false;

	protected boolean			isDestroy	= false;

	protected final SPL			storyPatternLink;

	protected final SPO[]		storyPatternObjects;

	protected final EMatchType	matchType;

	public MLSDMPatternPart(
			final PatternPartBasedMatcher<?, ?, ?, ?, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, ?, MLExpression> patternMatcher,
			final SPL storyPatternLink, final SPO[] storyPatternObjects)
	{
		super(patternMatcher);
		assert storyPatternObjects.length > 0;
		assert storyPatternLink != null || storyPatternObjects.length == 1;

		this.storyPatternLink = storyPatternLink;
		this.storyPatternObjects = storyPatternObjects;

		this.matchType = this.doGetMatchType();

		if (storyPatternLink != null)
		{
			switch (storyPatternLink.getModifier())
			{
				case CREATE:
					this.isCreate = true;
					break;

				case DESTROY:
					this.isDestroy = true;
					break;

				default:
			}
		}

		for (final SPO spo : storyPatternObjects)
		{
			switch (spo.getModifier())
			{
				case CREATE:
					this.isCreate = true;
					break;

				case DESTROY:
					this.isDestroy = true;
					break;

				default:
			}
		}
	}

	@Override
	public AbstractStoryPatternObject[] getStoryPatternObjects()
	{
		return this.storyPatternObjects;
	}

	public SPL getStoryPatternLink()
	{
		return this.storyPatternLink;
	}

	@Override
	public void createObjects()
	throws SDMException	{
		if (this.isCreate)
		{
			/*
			 * Create objects
			 */
			for (final AbstractStoryPatternObject spo : this.storyPatternObjects)
			{
				if (spo.getModifier() == ModifierEnum.CREATE)
				{
					final EClass spoEClass = spo.eClass();

					if (spoEClass == MlstorypatternsPackage.Literals.STORY_PATTERN_OBJECT)
					{
						/*
						 * Skip objects that already exist. This may be the case
						 * for optionally created objects and objects already
						 * created by another pattern part.
						 */
						if (!this.patternMatcher.getVariablesScope().variableExists(spo.getName()))
						{
							final StoryPatternObject storyPatternObject = (StoryPatternObject) spo;

							assert storyPatternObject.getType() instanceof EClass;
							assert !storyPatternObject.getType().eIsProxy();

							final EObject eObject;
							if(storyPatternObject.getAssignmentExpression() != null) {
								try {
									eObject = (EObject) this.patternMatcher.getExpressionInterpreterManager().evaluateExpression(storyPatternObject.getAssignmentExpression(), null,
																														null, this.patternMatcher.getVariablesScope()).getValue();
								} catch ( final ExpressionsInterpreterException ex ) {
									throw new SDMException( ex );
								}
							}
							else {
								final EFactory factory = storyPatternObject.getType().getEPackage().getEFactoryInstance();

								eObject = factory.create((EClass) storyPatternObject.getType());
							}
							
							this.patternMatcher.getVariablesScope().createVariable(storyPatternObject.getName(),
									storyPatternObject.getType(), eObject);

							this.patternMatcher.getNotificationEmitter().instanceObjectCreated(spo, eObject,
									this.patternMatcher.getVariablesScope(), this.patternMatcher);
						}
					}
					else
					{
						throw new UnsupportedOperationException();
					}
				}
			}
		}
	}

	@Override
	public void destroyObjects()
	{
		if (this.isDestroy)
		{
			for (final AbstractStoryPatternObject spo : this.storyPatternObjects)
			{
				if (spo.getModifier() == ModifierEnum.DESTROY)
				{
					final Variable<EClassifier> variable = this.patternMatcher.getVariablesScope().deleteVariable(spo.getName());

					/*
					 * Variable may be null in case of optional objects and
					 * objects destroyed in a different pattern part.
					 */
					if (variable != null)
					{
						assert variable.getValue() instanceof EObject;

						final EObject instanceObject = (EObject) variable.getValue();

						assert instanceObject != null;

						// Ticket # 2039 Optimization of EMF delete
						patternMatcher.getFacadeFactory().getInstanceFacade().delete( instanceObject );
						//EcoreUtil.delete(instanceObject, true);

						this.patternMatcher.getNotificationEmitter().instanceObjectDestroyed(spo, instanceObject,
								this.patternMatcher.getVariablesScope(), this.patternMatcher);
					}
				}
			}
		}
	}

	@Override
	public void createLinks() throws SDMException
	{
		if (this.isCreate)
		{
			this.doCreateLink();
		}
	}

	protected abstract void doCreateLink();

	@Override
	public final void destroyLinks(final java.util.Map<AbstractStoryPatternObject, Object> deletedObjects)
	{
		if (this.isDestroy)
		{
			this.doDestroyLink(deletedObjects);
		}

	};

	protected abstract void doDestroyLink(Map<AbstractStoryPatternObject, Object> deletedObjects);

	/**
	 * Matches targetSpo to targetObject. Performs several checks, e.g., for
	 * isomorphism, and emits notifications if the match was successful.
	 * 
	 * @param sourceSpo
	 * @param sourceInstanceObject
	 * @param targetSpo
	 * @param targetObject
	 * @return
	 * @throws SDMException
	 */
	@Override
	public EMatchType getMatchType()
	{
		return this.matchType;
	}

	private int	numberOfIncomingUnidirectionalLinks	= -1;

	public int getNumberOfIncomingUnidirectionalLinks(final AbstractStoryPatternObject spo)
	{
		if (this.numberOfIncomingUnidirectionalLinks == -1)
		{
			int i = 0;

			for (final AbstractStoryPatternLink spl : spo.getIncomingLinks())
			{
				if (!this.isReverseNavigable(spl))
				{
					i++;
				}
			}
			this.numberOfIncomingUnidirectionalLinks = i;
		}

		return this.numberOfIncomingUnidirectionalLinks;
	}

	protected boolean isReverseNavigable(final AbstractStoryPatternLink spl)
	{
		final EClass eClass = spl.eClass();

		if (eClass == MlstorypatternsPackage.Literals.CONTAINMENT_LINK)
		{
			return true;
		}
		else if (eClass == MlstorypatternsPackage.Literals.EXPRESSION_LINK)
		{
			return false;
		}
		else if (eClass == MlstorypatternsPackage.Literals.STORY_PATTERN_LINK)
		{
			final StoryPatternLink link = (StoryPatternLink) spl;

			if (link.getFeature() instanceof EReference)
			{
				final EReference ref = (EReference) link.getFeature();

				return ref.isContainment() || ref.getEOpposite() != null;
			}
			else
			{
				return false;
			}
		}
		else if (eClass == MlstorypatternsPackage.Literals.MAP_ENTRY_LINK)
		{
			return false;
		}
		else
		{
			throw new UnsupportedOperationException();
		}
	}

	protected abstract EMatchType doGetMatchType();
	
	@Override
	public String toString() {
		final StringBuilder strBuilder = new StringBuilder( getClass().getSimpleName() );
		strBuilder.append( " link " );
		strBuilder.append( String.valueOf( getStoryPatternLink() ) );
	
		return strBuilder.toString();
	}
}
