package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;

import de.mdelab.expressions.interpreter.core.ExpressionsInterpreterException;
import de.mdelab.expressions.interpreter.core.variables.Variable;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.MLStringExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlsdm.mlindices.Index;
import de.mdelab.mlsdm.mlindices.IndexEntry;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.icl.iCL.ICLConstraint;
import de.mdelab.sdm.icl.iCL.ICLPackage;
import de.mdelab.sdm.icl.iCL.ICLParameter;
import de.mdelab.sdm.icl.iCL.ICLRange;
import de.mdelab.sdm.icl.iCL.ICLStringExpression;
import de.mdelab.sdm.icl.iCL.ICLValue;
import de.mdelab.sdm.icl.iCL.ICLVariable;
import de.mdelab.sdm.icl.iCL.ICLVariableAttribute;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchMultipleNodesMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a generic superclass for matching actions based
 * on index constraints. It provides shared utility methods.
 * 
 * @author Matthias.Barkowsky
 *
 */
public abstract class MLSDMIndexMatchingAction extends
		MLSDMMatchingAction {

	/**
	 * This is a match state specific to the index staged
	 * matching action.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	protected static class MLSDMIndexEntryIteratorMatchState extends MatchState{
		
		/*
		 * Iterator of all matching index entries.
		 */
		private Iterator<IndexEntry> iterator;

		protected MLSDMIndexEntryIteratorMatchState(Iterator<IndexEntry> iterator) {
			this.iterator = iterator;
		}
		
		protected Iterator<IndexEntry> getEntryIterator() {
			return iterator;
		}
	}
	
	/*
	 * Mapping from ICL values in the query template of the
	 * ICL constraint to pattern nodes.
	 */
	protected Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNode;
	
	/*
	 * The ICL constraint.
	 */
	protected ICLConstraint iclConstraint;
	
	/*
	 * Object representing the ICL constraint in the story
	 * pattern.
	 */
	protected MLStringExpression expression;
	
	/*
	 * Index over which the constraint is formulated.
	 */
	protected Index index;
	
	public MLSDMIndexMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, ICLConstraint iclConstraint, MLStringExpression expression,
			Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNode) {
		super(id, matcher);
		this.iclValueToPatternNode = iclValueToPatternNode;
		this.iclConstraint = iclConstraint;
		this.expression = expression;
		this.index = (Index) matcher.getVariable(iclConstraint.getIndexID()).getValue();
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		List<Object> query = constructQuery();
		Iterator<IndexEntry> entryIterator = index.getEntries(query);
		return new MLSDMIndexEntryIteratorMatchState(entryIterator);
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		boolean checkOnly = true;
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dependency:getPatternConstraint().getDependencies()) {
			if(!dependency.isBound()) {
				checkOnly = false;
				break;
			}
		}
		if(checkOnly) {
			return PatternConstraint.CHECK_ONLY;
		}
		
		List<Object> query = constructQuery();
		int cardinality = index.estimateCardinality(query);
		
		/*
		 * If the query has no result for sure, this can be returned to allow early pruning.
		 * Otherwise only return the cardinality if the query is actually executable.
		 */
		return (cardinality == 0 || isExecutable()) ? cardinality : PatternConstraint.MATCHING_NOT_POSSIBLE;
	}

	/**
	 * Check whether the matching action can be executed, i.e.
	 * all required dependencies have already been bound.
	 * 
	 * @return	True if the action is executable.
	 */
	protected abstract boolean isExecutable();
	
	/**
	 * Create an index query based on which dependencies
	 * are bound and which are not.
	 * 
	 * @return	An index query conforming to the current
	 * 			state of the search.
	 */
	protected abstract List<Object> constructQuery();

	/**
	 * Retrieve the concrete value represented by an ICL value
	 * in the query template in the current context.
	 * 
	 * @param value
	 * 			The ICL value for which to retrieve the concrete
	 * 			value.
	 * @return	The concrete value.
	 */
	protected Object getParameterValue(ICLValue value) {
		if(iclValueToPatternNode.containsKey(value)) {
			for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode:iclValueToPatternNode.get(value)) {
				if(!patternNode.isBound()) {
					return null;
				}
			}
		}
		
		if(value.eClass() == ICLPackage.Literals.ICL_VARIABLE_ATTRIBUTE) {
			return getVariableAttributeValue((ICLVariableAttribute) value);
		}
		else if(value.eClass() == ICLPackage.Literals.ICL_VARIABLE) {
			return getVariableValue((ICLVariable) value);
		}
		else if (value.eClass() == ICLPackage.Literals.ICL_STRING_EXPRESSION) {
			return getStringExpressionValue((ICLStringExpression) value);
		}
		else {
			return Index.UNDEFINED_PARAMETER;
		}
	}

	/**
	 * Retrieve the concrete value represented by an
	 * ICL variable.
	 * 
	 * @param value
	 * 			The ICL value for which to retrieve the concrete
	 * 			value.
	 * @return	The concrete value.
	 */
	private Object getVariableValue(ICLVariable value) {
		Variable<EClassifier> variable = matcher.getVariable(((ICLVariable) value).getName());
		if(variable == null) {
			return null;
		}
		else{
			return variable.getValue();
		}
	}

	/**
	 * Retrieve the concrete value represented by an
	 * ICL variable attribute.
	 * 
	 * @param variableAttribute
	 * 			The ICL value for which to retrieve the concrete
	 * 			value.
	 * @return	The concrete value.
	 */
	private Object getVariableAttributeValue(ICLVariableAttribute variableAttribute) {
		Variable<EClassifier> variable = matcher.getVariable(variableAttribute.getVariable().getName());
		if(variable == null) {
			return null;
		}
		else{
			EObject eObject = (EObject) variable.getValue();
			return eObject == null ? null : eObject.eGet(eObject.eClass().getEStructuralFeature(variableAttribute.getAttribute()));
		}
	}
	
	/**
	 * Retrieve the concrete value represented by an
	 * ICL string expression.
	 * 
	 * @param value
	 * 			The ICL value for which to retrieve the concrete
	 * 			value.
	 * @return	The concrete value.
	 */
	private Object getStringExpressionValue(ICLStringExpression value) {
		try {
			return matcher.getExpressionInterpreterManager().evaluateExpression(((ICLStringExpression) value).getExpression(), EcorePackage.Literals.EOBJECT, null, matcher.getVariablesScope()).getValue();
		} catch (ExpressionsInterpreterException e) {
			e.printStackTrace();
		}
		return Index.UNDEFINED_PARAMETER;
	}

	protected boolean checkValue(ICLValue value, Object object) {
		Object parameterValue = getParameterValue(value);
		return parameterValue == Index.UNDEFINED_PARAMETER || parameterValue.equals(object);
	}
	
	/**
	 * Check whether an object matches a range specified
	 * in an index query template.
	 * 
	 * @param range
	 * 			The specified range.
	 * @param object
	 * 			The object to check.
	 * @return	True if the object falls into the specified
	 * 			range.
	 */
	@SuppressWarnings({ "unchecked", "rawtypes" })
	protected boolean checkRange(ICLRange range, Object object) {
		Object min = getParameterValue(range.getMin());
		Object max = getParameterValue(range.getMax());
		if(!(object instanceof Comparable)) {
			return false;
		}
		else {
			return (min == Index.UNDEFINED_PARAMETER || ((Comparable) object).compareTo(min) >= 0) &&
					(max == Index.UNDEFINED_PARAMETER || ((Comparable) object).compareTo(max) <= 0);
		}
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		MatchMultipleNodesMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new MatchMultipleNodesMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
		transaction.setValid(false);
		
		if(!(matchState instanceof MLSDMIndexEntryIteratorMatchState)) {
			return transaction;
		}
		
		matcher.getNotificationEmitter().evaluatingExpression(expression, matcher.getVariablesScope(), matcher);
		
		Iterator<IndexEntry> iterator = ((MLSDMIndexEntryIteratorMatchState) matchState).getEntryIterator();
		boolean valid = false;
		
		while(!valid && iterator.hasNext()) {
			transaction = new MatchMultipleNodesMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
			IndexEntry entry = iterator.next();
			
			/*
			 * If entry has already been used for matching skip to the next one.
			 */
			if(!matcher.checkMatchingHistory(this.getPatternConstraint(), entry)) {
				valid = false;
				transaction.setValid(valid);
				continue;
			}
			
			/*
			 * Check if the current entry is valid and create new bindings.
			 */
			List<ICLParameter> parameters = iclConstraint.getParameters().getList();
			Set<Object> newlyBoundObjects = new HashSet<Object>();
			valid = true;
			
			for(int i = 0; i < parameters.size(); i++) {
				ICLParameter parameter = parameters.get(i);
				Object indexValue = entry.getKey().get(i);

				if(parameter.eClass() == ICLPackage.Literals.ICL_VARIABLE &&
						iclValueToPatternNode.containsKey(parameter)) {
					for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode:iclValueToPatternNode.get(parameter)) {
						if(patternNode.isBound()) {
							if(!checkValue((ICLValue) parameter, indexValue)) {
								valid = false;
								break;
							}
							else {
								continue;
							}
						}
						/*
						 * Create subtransactions for each previously unbound Pattern Node.
						 */
						
						if(matcher.instanceMatchesSPO(indexValue, patternNode.getSpo()) && !newlyBoundObjects.contains(indexValue)){
							MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> singleTransaction =
									new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
							singleTransaction.setValid(true);
							singleTransaction.setSPO(patternNode.getSpo());
							singleTransaction.setTargetInstance(indexValue);
							
							transaction.addSingleNodeTransaction(singleTransaction);
							transaction.addPatternNode(patternNode);
							
							newlyBoundObjects.add(indexValue);
						}
						else {
							valid = false;
							break;
						}
					}
				}
				else {
					/*
					 * Check already bound parameter.
					 */
					if(parameter.eClass() == ICLPackage.Literals.ICL_RANGE) {
						if(!checkRange((ICLRange) parameter, indexValue)) {
							valid = false;
							break;
						}
					}
					else {
						if(!checkValue((ICLValue) parameter, indexValue)) {
							valid = false;
							break;
						}
					}
				}
			}
			
			matcher.updateMatchingHistory(this.getPatternConstraint(), entry);
			transaction.setValid(valid);
		}
		
		return transaction;
	}

}
