package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for matching the source node
 * of a link in the story pattern. It can be executed once
 * the target node of the link has been mapped. It uses the
 * data structures provided by a reference adapter.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMLinkTraverseReverseMatchingAction extends MLSDMMatchingAction {

	/*
	 * The link.
	 */
	private StoryPatternLink link;
	
	/*
	 * The link type.
	 */
	private EStructuralFeature feature;
	
	/*
	 * Source node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode;
	
	/*
	 * Target node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode;
	
	/**
	 * This is a match state specific to the traverse reverse
	 * link matching action.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	private static class MLSDMTraverseReverseLinkMatchState extends MatchState{

		private Iterator<? extends Object> featureIterator;
		
		private MLSDMTraverseReverseLinkMatchState(Iterator<? extends Object> featureIterator) {
			this.featureIterator = featureIterator;
		}

		private Iterator<? extends Object> getFeatureIterator() {
			return featureIterator;
		}
		
	}
	
	public MLSDMLinkTraverseReverseMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, StoryPatternLink link,
			EStructuralFeature feature,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode) {
		super(id, matcher);
		this.link = link;
		this.feature = feature;
		this.sourceNode = sourceNode;
		this.targetNode = targetNode;
		addRequiredBinding(targetNode);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality()  {
		AbstractStoryPatternObject targetSPO = targetNode.getSpo();
		
		if(!sourceNode.isBound() && targetNode.isBound()) {
			Object targetInstance = matcher.getVariablesScope().getVariable(targetSPO.getName()).getValue();
			
			return matcher.getReferenceAdapter().getInverseReferences((EObject) targetInstance, (EReference) feature).size();
		}
		else {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		AbstractStoryPatternObject targetSPO = targetNode.getSpo();
		if(sourceNode.isBound() || !targetNode.isBound()) {
			return null;
		}

		EObject targetObject = (EObject) matcher.getVariable(targetSPO.getName()).getValue();
		Iterator<? extends Object> instanceLinkIterator = matcher.getReferenceAdapter().getInverseReferences(targetObject, (EReference) feature).iterator();
		
		return new MLSDMTraverseReverseLinkMatchState(instanceLinkIterator);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
		
		EObject targetInstance = (EObject) matcher.getVariable(targetNode.getSpo().getName()).getValue();
		matcher.getNotificationEmitter().traversingLink(link, link.getSource(), targetInstance, link.getTarget(), matcher.getVariablesScope(), matcher);
		
		if(!(matchState instanceof MLSDMTraverseReverseLinkMatchState)) {
			transaction.setValid(false);
		}
		else {
			boolean valid = false;
			MLSDMTraverseReverseLinkMatchState ms = (MLSDMTraverseReverseLinkMatchState) matchState;
			Iterator<? extends Object> featureIterator = ms.getFeatureIterator();
			EObject sourceInstance = null;
			while(!valid && featureIterator.hasNext()) {
				sourceInstance = (EObject) featureIterator.next();
				valid = 	matcher.instanceMatchesSPO(sourceInstance, sourceNode.getSpo())
							&& matcher.checkMatchingHistory(this.getPatternConstraint(), sourceInstance)
							&& checkPositionConstraint(sourceInstance, targetInstance);
			}
			
			if(valid) {
				transaction.setTargetInstance(sourceInstance);
				transaction.setSPO(sourceNode.getSpo());
			}
			else {
				matcher.getNotificationEmitter().storyPatternObjectNotBound(sourceNode.getSpo(), matcher.getVariablesScope(),
						matcher);
			}
			
			transaction.setValid(valid);
			transaction.addPatternNode(sourceNode);

			matcher.updateMatchingHistory(this.getPatternConstraint(), sourceInstance);
		}
		return transaction;
	}
	
	
	/**
	 * Check whether the given instance objects satisfy the position
	 * constraint imposed over this link.
	 * 
	 * @param e1
	 * 			the source instance
	 * @param e2
	 * 			the target instance
	 * @return	true if the position constraint is satisfied
	 */
	private boolean checkPositionConstraint(EObject e1, EObject e2) {
		List<?> instanceLinks;
		switch(link.getLinkPositionConstraint()) {
			case BETWEEN :
				instanceLinks = (List<?>)e1.eGet(link.getFeature());
				int index = instanceLinks.indexOf(e2);
				return index > 1 && index < instanceLinks.size() - 1;
			case UNCONSTRAINED :
				return true;
			case FIRST :
				instanceLinks = (List<?>)e1.eGet(link.getFeature());
				return instanceLinks.get(0) == e2;
			case LAST :
				instanceLinks = (List<?>)e1.eGet(link.getFeature());
				return instanceLinks.get(instanceLinks.size() - 1) == e2;
			default :
				return true;
		}
	}

	public EStructuralFeature getFeature() {
		return feature;
	}

	public PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getSourceNode() {
		return sourceNode;
	}

	public PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getTargetNode() {
		return targetNode;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "traverse\t(" + link.getTarget().getName() + " <-" + link.getFeature().getName() + "- " + link.getSource().getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMLinkTraverseReverseMatchingAction(id, matcher, link, feature, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(sourceNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(targetNode));
	}

}
