package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.ExpressionsInterpreterException;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.MLStringExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckOnlyMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for executing an explicit
 * check of an expression constraint formulated over a
 * single node in the story pattern.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMSPOExpressionCheckMatchingAction extends MLSDMCheckMatchingAction {

	/*
	 * Object representing the expression constraint
	 * in the story pattern.
	 */
	private MLExpression expression;
	
	/*
	 * Pattern node over which the constraint is
	 * formulated.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode;
	
	public MLSDMSPOExpressionCheckMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher,
			MLExpression expression, PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode) {
		super(id, matcher);
		this.patternNode = patternNode;
		this.expression = expression;
		addRequiredBinding(patternNode);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality()  {
		if(patternNode.isBound()) {
			return PatternConstraint.CHECK_ONLY;
		}
		else {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
	}

	/**
	 * Perform the actual check by evaluating the expression
	 * in the current context using the expression interpreter
	 * manager.
	 * 
	 * @param contextClassifier
	 * 			The type of the node over which the constraint
	 * 			is formulated.
	 * @param instanceObject
	 * 			The instance object to which the node is currently
	 * 			mapped.
	 * @return
	 */
	private boolean checkExpression(EClassifier contextClassifier, Object instanceObject) {
		boolean success = false;
		try {
			success = (Boolean) matcher.getExpressionInterpreterManager().evaluateExpression(expression, contextClassifier, instanceObject, matcher.getVariablesScope()).getValue();
		} catch (ExpressionsInterpreterException e) {
			e.printStackTrace();
		}
		

		if(success) {
			matcher.getNotificationEmitter().storyPatternConstraintHolds(expression, matcher.getStoryPattern(), matcher.getVariablesScope(), matcher);
		}
		else {
			matcher.getNotificationEmitter().storyPatternConstraintViolated(expression, matcher.getStoryPattern(), matcher.getVariablesScope(), matcher);
		}
		
		return success;
	}
	
	public MLExpression getExpression() {
		return expression;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		return new MLSDMCheckOnceMatchState();
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
		if(!(matchState instanceof MLSDMCheckOnceMatchState)) {
			transaction.setValid(false);
		}
		else {
			transaction.setValid(((MLSDMCheckOnceMatchState) matchState).check() && checkExpression(patternNode.getSpo().getType(), matcher.getVariable(patternNode.getSpo().getName()).getValue()));
		}
		return transaction;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckMatchingAction#check()
	 */
	@Override
	public boolean check() {
		return checkExpression(patternNode.getSpo().getType(), matcher.getVariable(patternNode.getSpo().getName()).getValue());
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "check\t\t(" + patternNode.getSpo().getName() + ": " + (expression instanceof MLStringExpression ? ((MLStringExpression) expression).getExpressionString() : expression.toString()) + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMSPOExpressionCheckMatchingAction(id, matcher, expression, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(patternNode));
	}
}
