package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher;

import java.util.Collections;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.ExpressionInterpreterManager;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.interpreter.MLSDMExpressionInterpreterManager;
import de.mdelab.mlsdm.interpreter.facade.MLSDMMetamodelFacadeFactory;
import de.mdelab.mlsdm.interpreter.notifications.MLSDMNotificationEmitter;
import de.mdelab.mlsdm.interpreter.searchModel.MLSDMSearchModelBasedInterpreter;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.MLSDMStrategyFactory;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.facade.MetamodelFacadeFactory;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;
import de.mdelab.sdm.interpreter.core.patternmatcher.MatchApplier;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelBasedMatcher;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelBuilder;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzerManager;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.PatternConstraintSelectionStrategy;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

/**
 * This is an implementation of the search model based matcher for the
 * MLSDM meta model.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMSearchModelBasedMatcher extends SearchModelBasedMatcher<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> {

	private static MLSDMStrategyFactory DEFAULT_FACTORY = new MLSDMStrategyFactory();
	
	/*
	 * Reference adapter for optimized edge matching.
	 */
	protected MLSDMReferenceIndex referenceAdapter;
	
	protected MLSDMStrategyFactory strategyFactory;

	public MLSDMSearchModelBasedMatcher(StoryPattern storyPattern,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope,
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> expressionAnalyzerManager,
			MLSDMReferenceIndex referenceAdapter,
			final MLSDMNotificationEmitter notificationEmitter ) throws SDMException {
		this(storyPattern, variablesScope, MLSDMMetamodelFacadeFactory.INSTANCE, expressionInterpreterManager, expressionAnalyzerManager, DEFAULT_FACTORY, referenceAdapter, notificationEmitter, Collections.<String, Object>emptyMap(), new MLSDMSearchModelBuilder());
	}

	public MLSDMSearchModelBasedMatcher(StoryPattern storyPattern,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope,
			MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory,
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> expressionAnalyzerManager,
			MLSDMReferenceIndex referenceAdapter,
			final MLSDMNotificationEmitter notificationEmitter) throws SDMException {
		this(storyPattern, variablesScope, facadeFactory, expressionInterpreterManager, expressionAnalyzerManager, DEFAULT_FACTORY, referenceAdapter, notificationEmitter, Collections.<String, Object>emptyMap(), new MLSDMSearchModelBuilder());
	}

	public MLSDMSearchModelBasedMatcher(StoryPattern storyPattern,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope,
			MLSDMMetamodelFacadeFactory facadeFactory,
			MLSDMExpressionInterpreterManager expressionInterpreterManager,
			ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> expressionAnalyzerManager,
			MLSDMStrategyFactory strategyFactory,
			MLSDMReferenceIndex referenceAdapter) throws SDMException {
		this(storyPattern, variablesScope, facadeFactory, expressionInterpreterManager, expressionAnalyzerManager, strategyFactory, referenceAdapter, expressionInterpreterManager.getNotificationEmitter(), Collections.<String, Object>emptyMap(), new MLSDMSearchModelBuilder());
	}

	public MLSDMSearchModelBasedMatcher(StoryPattern storyPattern,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope,
			MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory,
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> expressionAnalyzerManager,
			MLSDMReferenceIndex referenceAdapter,
			NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> notificationEmitter,
			Map<String, Object> interpreterParameters) throws SDMException {
		this(storyPattern, variablesScope, facadeFactory, expressionInterpreterManager, expressionAnalyzerManager,
				interpreterParameters.containsKey(MLSDMSearchModelBasedInterpreter.STRATEGY_FACTORY) ? (MLSDMStrategyFactory) interpreterParameters.get(MLSDMSearchModelBasedInterpreter.STRATEGY_FACTORY) : DEFAULT_FACTORY,
				referenceAdapter, notificationEmitter, interpreterParameters, new MLSDMSearchModelBuilder());
	}
	
	public MLSDMSearchModelBasedMatcher(StoryPattern storyPattern,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope,
			MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory,
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> expressionAnalyzerManager,
			MLSDMStrategyFactory strategyFactory,
			MLSDMReferenceIndex referenceAdapter,
			NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> notificationEmitter) throws SDMException {
		this(storyPattern, variablesScope, facadeFactory, expressionInterpreterManager, expressionAnalyzerManager, strategyFactory, referenceAdapter, notificationEmitter, Collections.<String, Object>emptyMap(), new MLSDMSearchModelBuilder());
	}
	
	public MLSDMSearchModelBasedMatcher(StoryPattern storyPattern,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope,
			MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory,
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> expressionAnalyzerManager,
			MLSDMStrategyFactory strategyFactory,
			MLSDMReferenceIndex referenceAdapter,
			NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> notificationEmitter,
			Map<String, Object> interpreterParameters) throws SDMException {
		this(storyPattern, variablesScope, facadeFactory, expressionInterpreterManager, expressionAnalyzerManager, strategyFactory, referenceAdapter, notificationEmitter, interpreterParameters, new MLSDMSearchModelBuilder());
	}
	
	public MLSDMSearchModelBasedMatcher(StoryPattern storyPattern,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope,
			MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory,
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> expressionAnalyzerManager,
			MLSDMStrategyFactory strategyFactory,
			MLSDMReferenceIndex referenceAdapter,
			NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> notificationEmitter,
			Map<String, Object> interpreterParameters,
			SearchModelBuilder<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModelBuilder) throws SDMException {
		super(storyPattern, variablesScope, facadeFactory, expressionInterpreterManager,
				expressionAnalyzerManager, strategyFactory.createMatchingStrategy(storyPattern, referenceAdapter, variablesScope), notificationEmitter, interpreterParameters, searchModelBuilder);
		this.referenceAdapter = referenceAdapter;
		this.strategyFactory = strategyFactory;
		this.matchApplier = createMatchApplier(storyPattern, variablesScope, facadeFactory, expressionInterpreterManager, notificationEmitter);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelBasedMatcher#createMatchApplier(java.lang.Object, de.mdelab.sdm.interpreter.core.variables.VariablesScope, de.mdelab.sdm.interpreter.core.facade.MetamodelFacadeFactory, de.mdelab.sdm.interpreter.core.expressions.ExpressionInterpreterManager, de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MutableNotificationEmitterWrapper)
	 */
	@Override
	protected MatchApplier<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> createMatchApplier(
			StoryPattern storyPattern,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope,
			MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory,
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> notificationEmitter) throws SDMException {
		return new MLSDMSearchModelBasedMatchApplier(storyPattern, facadeFactory, expressionInterpreterManager, notificationEmitter, referenceAdapter);
	}

	public MLSDMReferenceIndex getReferenceAdapter() {
		return referenceAdapter;
	}

	@Override
	protected PatternConstraintSelectionStrategy<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> createSelectionStrategy() {
		return strategyFactory.createMatchingStrategy(getStoryPattern(), referenceAdapter, getVariablesScope());
	}

	public MLSDMStrategyFactory getStrategyFactory() {
		return strategyFactory;
	}

}
