package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.Adornment;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.CostModel;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModel;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

public class MLSDMUpdatingHybridCostStrategy
		extends
			MLSDMHybridCostStrategy {
	
	double[] domainSizes;
	
	public MLSDMUpdatingHybridCostStrategy(
			MLSDMHybridSelectionStrategy selectionStrategy,
			SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel,
			CostModel costModel,
			NotifierVariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope,
			MLSDMMetadataIndex metadataAdapter) {
		super(selectionStrategy, searchModel, costModel, variablesScope, metadataAdapter);
	}

	@Override
	public void initialize() {
		super.initialize();
		initializeDomainSizes();
	}
	
	private void initializeDomainSizes() {
		int maxId = 0;
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc:searchModel.getPatternConstraints()) {
			for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma:pc.getPossibleActions()) {
				maxId = Math.max(maxId, ma.getId() + 1);
			}
		}
		domainSizes = new double[maxId];
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc:searchModel.getPatternConstraints()) {
			for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma:pc.getPossibleActions()) {
				domainSizes[ma.getId()] = getRemainingDomainSize(ma);
			}
		}
	}

	@Override
	public double calculateMatchingCost(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint) {
		if(patternConstraint.getCardinality() == PatternConstraint.CHECK_ONLY) {
			return PatternConstraint.CHECK_ONLY;
		}
		
		int oldBindingCode = selectionStrategy.getCurrentBinding();
		int newBindingCode = oldBindingCode;
		int newBindings = 0;
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:patternConstraint.getDependencies()) {
			if(!node.isBound()) {
				newBindingCode = newBindingCode | (1 << selectionStrategy.getIndex(node));
				newBindings++;
			}
		}
		
		double constraintCardinality = patternConstraint.getCardinality();
		double filtering = getFiltering(continuationTable[oldBindingCode], continuationTable[newBindingCode], patternConstraint);
		return 	(constraintCardinality * newBindings) +
				(constraintCardinality * filtering * getUpdatedTotalCost(continuationTable[newBindingCode]));
	}

	private double getFiltering(Adornment oldAdornment, Adornment newAdornment,
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint) {
		double filtering = 1;
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc:oldAdornment.matchingOrder) {
			if(pc != patternConstraint && !newAdornment.matchingOrder.contains(pc)) {
				filtering *= getBestFilteringEstimate(pc);
			}
		}
		return filtering;
	}

	private double getUpdatedTotalCost(Adornment adornment) {
		double cost = 0;
		double front = 1;
		for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma:adornment.actions) {
			if(ma.isCheck()) {
				front *= getBestFilteringEstimate(ma.getPatternConstraint());
			}
			else {
				int newBindings = (ma.getPatternConstraint().getDependencies().size() - ma.getRequiredBindings().size());
				double estimate = getBestCostEstimate(ma);
				cost += front * newBindings * estimate;
				front *= estimate;	
			}
		}
		return cost;
	}

	private double getBestCostEstimate(
			MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		if(ma != ma.getPatternConstraint().getActiveAction()) {
			return costCalculator.estimateCost(ma);
		}
		else {
			return ma.getPatternConstraint().getCardinality();
		}
	}

	private double getBestFilteringEstimate(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc) {
		if(pc.getActiveAction() == null) {
			return costCalculator.getFiltering(pc.getExplicitCheckAction());
		}
		else {
			return pc.getCardinality() / domainSizes[pc.getActiveAction().getId()];
		}
	}

	private double getRemainingDomainSize(
			MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		double size = 1;
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn:ma.getPatternConstraint().getDependencies()) {
			if(!ma.getRequiredBindings().contains(pn)) {
				size *= getDomainSize(pn);
			}
		}
		return size;
	}

	private double getDomainSize(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn) {
		return this.selectionStrategy.metadataAdapter.getDomain((EClass) pn.getSpo().getType()).size();
	}

}
